unit BaseTabModalFra;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, TabClientFra, StdCtrls, Contnrs, Generics.Collections;

type
  TfraBaseTabModal = class(TFrame)
    pnlCaption: TPanel;
    btnClose: TButton;
    procedure btnCloseClick(Sender: TObject);
  private
    { Private-Deklarationen }
    FModalResult: TModalResult;
  protected
    function CloseQuery: Boolean; virtual;
    procedure DoClose(ACloseAction: TCloseAction); virtual;
    procedure CloseModal;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public-Deklarationen }
    function ShowTabModal(AClient: TfraTabClient): TModalResult;

    property ModalResult: TModalResult read FModalResult write FModalResult;
  end;

procedure SwitchToModalMessageLoop;

implementation

{$R *.dfm}

var
  MainFiber: Pointer;
  FiberToDelete: Pointer;
  TabModalFrames: TDictionary<TfraBaseTabModal, Pointer>;

function GetCurrentFiber: Pointer;
asm
  mov eax, fs:[$10]
end;

procedure ReturnToFiber(AFiber: Pointer);
begin
  FiberToDelete := GetCurrentFiber;
  SwitchToFiber(AFiber);
end;

procedure ModalMessageLoop(Param: LPARAM); stdcall;
var
  ReturnFiber: Pointer;
  Item: TPair<TfraBaseTabModal, Pointer>;
begin
  ReturnFiber := MainFiber;
  repeat
    try
      if TabModalFrames <> nil then
      begin
        for Item in TabModalFrames do
        begin
          if Item.Key.Enabled then
          begin
            if Application.Terminated then
              Item.Key.ModalResult := mrCancel
            else if Item.Key.ModalResult <> mrNone then
              Item.Key.CloseModal;

            if Item.Key.ModalResult <> mrNone then
            begin
              ReturnFiber := Item.Value;
              ReturnToFiber(ReturnFiber);
            end;
          end;
        end;
      end;
      Application.HandleMessage;
    except
      Application.HandleException(nil);
    end;
  until Application.Terminated;
  ReturnToFiber(ReturnFiber);
end;

procedure SwitchToModalMessageLoop;
begin
  if MainFiber = nil then
    MainFiber := GetCurrentFiber;

  SwitchToFiber(CreateFiber(0, @ModalMessageLoop, nil));
  if FiberToDelete <> nil then
  begin
    DeleteFiber(FiberToDelete);
    FiberToDelete := nil;
  end;
end;

function DisableTabModalControls(AControl: TWinControl; AExcludeChild: TControl): TList;
var
  I: Integer;
begin
  Result := TList.Create;
  try
    for I := 0 to AControl.ControlCount - 1 do
      if (AControl.Controls[I] <> AExcludeChild) and AControl.Controls[I].Enabled then
      begin
        Result.Add(AControl.Controls[I]);
        AControl.Controls[I].Enabled := False;
      end;
  except
    Result.Free;
    raise;
  end;
end;

procedure EnableTabModalControls(AControl: TWinControl; var AList: TList);
var
  I: Integer;
begin
  for I := 0 to AControl.ControlCount - 1 do
    if not AControl.Controls[I].Enabled and (AList.IndexOf(AControl.Controls[I]) <> -1) then
      AControl.Controls[I].Enabled := True;
  FreeAndNil(AList);
end;

{ TfraBaseTabModal }

procedure TfraBaseTabModal.btnCloseClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TfraBaseTabModal.CloseModal;
var
  CloseAction: TCloseAction;
begin
  try
    CloseAction := caNone;
    if CloseQuery then
    begin
      CloseAction := caHide;
      DoClose(CloseAction);
    end;
    case CloseAction of
      caNone: ModalResult := mrNone;
      //caFree: Release;
    end;
  except
    ModalResult := mrNone;
    Application.HandleException(Self);
  end;
end;

function TfraBaseTabModal.CloseQuery: Boolean;
begin
  Result := True;
end;

procedure TfraBaseTabModal.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or WS_DLGFRAME;
end;

procedure TfraBaseTabModal.DoClose(ACloseAction: TCloseAction);
begin
end;

function TfraBaseTabModal.ShowTabModal(AClient: TfraTabClient): TModalResult;
var
  OldParent: TWinControl;
  DisabledList: TList;
begin
  Visible := False;
  OldParent := Parent;
  try
    Parent := AClient.GetModalParent;
    Left := (Parent.ClientWidth - Width) div 2;
    Top := (Parent.ClientHeight - Height) div 2;

    DisabledList := DisableTabModalControls(Parent, Self);
    try
      if TabModalFrames = nil then
        TabModalFrames := TDictionary<TfraBaseTabModal, Pointer>.Create;
      try
        TabModalFrames.Add(Self, GetCurrentFiber);
        Visible := True;
        try
          ModalResult := mrNone;
          SwitchToModalMessageLoop; // returns after tab-modal frame was closed
        finally
          Visible := False;
        end;
      finally
        TabModalFrames.Remove(Self);
        if TabModalFrames.Count = 0 then
          FreeAndNil(TabModalFrames);
      end;
    finally
      EnableTabModalControls(Parent, DisabledList);
    end;
  finally
    Parent := OldParent;
  end;
  Result := ModalResult;
end;

end.
